
package HostingMaster::Server;


=head1 HostingMaster::Server

Run preconfigured commands for remote users.
Password login and session identification based on varying challenge.

Lower level commands are saved from the need for synchronization
by only serving one client at a time.

Usage: new()->run()

=cut


use strict;

use IO::File;
use IO::Socket;
use IPC::Open3;
use Digest::MD5 qw(md5_base64);
use vars qw( @default $daemon );


=head1 Parameters

C<@default> contains default server object properties, constructed by
evaluation. These are overridable as named parameters to C<new()>.

Multiple values are separated by '|' pipe characters.

  daemonname       name of daemon binary (default from $0)
  passwordfiles    names of password files (default: /etc/daemonname.pwd)
  conffile         name of configuration file (/etc/daemonname.conf)
  logfile          name of logfile (default: /var/log/daemonname.log)
  pidfile          name of pidfile (default: /var/run/daemonname.pid)

  addresses        acceptable client IP addresses (default: all local)
  port             number of port to listen at
  session_timeout  seconds before session expires (default: 600)
  
Reserved runtime variables:

  open_sessions    hash of open session objects
  port_reserved    set when listening fails
  client_address   IP address of current client
  socket           hole in the wall to talk through with the client
  username         name of identified user

  users            hash of users and associated commands

  sigpipe          whether a pipe has been broken
  kill_on_sigpipe  id of process to terminate upon client exit

=cut


@default = 
    (
     'daemonname' =>    q{ ( $0 =~ /([^\/]+)$/ )[0] }, # file part of $0
     'passwordfiles' => q{ "/etc/$self->{daemonname}.pwd" },
     'conffile' =>      q{ "/etc/$self->{'daemonname'}.conf" },
     'logfile' =>       q{ "/var/log/$self->{'daemonname'}.log" },
     'pidfile' =>       q{ "/var/run/$self->{'daemonname'}.pid" },

     'addresses' =>     q{ join('|', (`/sbin/ifconfig -a` =~ 
				      /\binet (?:addr:|)([0-9.]+)/gis)) },
     'port' =>          q{ 31337 },
     'session_timeout' => q{ 600 },

     'open_sessions' => q{ {} },
     'port_reserved' => q{ 0 },
     'debug' => q{ 0 },
     );

=head1 new

  my $daemon = HostingMaster::Server->new( 'port' => $port, ... );
  my $daemon = HostingMaster::Server->new( { ... } );

  Constructs and configures a new server object.

=cut


sub new {
    my $class = ref($_[0]) || shift;
    my $self = { ref($_[0]) ? %{$_[0]} : @_ };

    my @eval = @default;
    while( my ($key, $eval) = splice(@eval, 0, 2) ) {
	unless( exists $self->{$key} ) {
	    $self->{$key} = eval $eval;
	    die "Failed to set $key: $@" if $@;
	}
    }

    bless $self, $class;

    if( $self->{'logfile'} ) {
	open STDERR, ">>$self->{'logfile'}"
	    or die "Failed to open logfile $self->{'logfile'}: $!\n";
    }

    $self->conf();

    return $self;
}


### daemonic behavior


=head1 wlog

Write a log entry in a nice format.

=cut


sub wlog {
    my $self = ref $_[0] ? shift : $daemon;
    chomp( my $msg = "@_" );
    my @time = (localtime())[0..5];
    $time[5] += 1900;
    $time[4] += 1;
    printf( STDERR "[%04d:%02d:%02d %02d:%02d:%02d] %d %s: %s\n",
	    reverse(@time), $$, $self->{'daemonname'}, $msg );
}


=head1 daemonize

Creates pidfile and grabs signals:

  TERM => death()
  PIPE => stand()
  HUP  => conf()

=cut


sub daemonize {
    my $self = shift;

    die "Already daemonized\n" if $daemon;
    $daemon = $self;

    my $pidf = $self->{'pidfile'};
    if( -f $pidf ) {
	chomp( my $rpid = `cat $pidf` );
	die "Already running ($pidf:$rpid)\n" if -d "/proc/$rpid";
    }
    my $pidfile = IO::File->new(">$pidf")
	or die "Failed to create pid file $pidf\n";
    print $pidfile "$$\n";
    close $pidfile;
    $self->{'mainpid'} ||= $$;

    $self->{'sigpipe'} = $self->{'kill_on_sigpipe'} = 0;

    $SIG{'TERM'} = \&death;
    $SIG{'PIPE'} = \&stand;
    $SIG{'HUP'} = \&conf;

    wlog "listen $self->{'addresses'}:$self->{'port'}\n";
}


=head1 death

Handles cleanup upon end of mother process.

=cut


sub death {
    if( $$ == $daemon->{'mainpid'} ) {
      wlog "server stopped.";
      unlink $daemon->{'pidfile'};
    }
    exit 0;
}


=head1 stand

Handles Pipe signal.

=cut


sub stand {
    wlog "caught signal @_ ($daemon->{'kill_on_sigpipe'})\n"
	if $daemon->{'debug'};
    $daemon->{'sigpipe'}++;
    kill 15, $daemon->{'kill_on_sigpipe'} if $daemon->{'kill_on_sigpipe'};
    $daemon->{'kill_on_sigpipe'} = 0;
}


=head1 conf

Reads configuration from file.

=cut


sub conf {
    my $self = ref $_[0] ? shift : $daemon;
    my ( %groups, %users );
    return unless -f $self->{'conffile'};

    my $conf = `cat '$self->{'conffile'}'`;
    eval $conf;
    die "$0 conf: $@" if $@;

    $self->{'users'} = \%users;

    $self->wlog( ref($self), " configured:",
		 map( { my $s=$self->{$_}; 
			" $_=" . (ref $s eq 'HASH'? keys %{$s}: $s ) }
		      sort keys %$self ), "\n" );
}


=head1 headers

Reads "name: value" header lines from given stream into a hash.
Space-prefixed values are appended to last name.

=cut


sub headers {
    my $self = shift;
    my $in = $self->{'socket'};
    my %headers;
    my $key = undef;
    my $i = 0;
    while( <$in> ) {
	die "Client disconnected\n" if $self->{'sigpipe'};
	chomp;
	last if /^\s*$/;
	$i++;
	if( /^\s+(.+)$/ and $key ) {
	    $headers{$key} .= $1;
	} elsif( /^(\w+)\s*[:=]\s*(.*)\s*$/ ) {
	    $key = lc($1);
	    $headers{$key} = $2;
	} else {
	    die "Bad header line $i.\n";
	}
    }
    return \%headers;
}


### identification


=head1 sessionid

Generates a random identification string.

=cut


my @sidchars = ('A'..'Z','a'..'z','0'..'9','_','-');
sub sessionid {
    return join '', map { $sidchars[int rand @sidchars] } (1..16);
}


=head1 challenge

Challenge-response dialog with client to ensure user identity.
Returns an identification seed.

=cut


sub challenge {
    my ( $self, $headers ) = @_;
    my $client = $self->{'socket'};
    my $user = $headers->{'username'};
    my ( $pwd, $seed );

    wlog "Challenge $user for identification\n";

    foreach my $pwdfile ( split /\|/, $self->{'passwordfiles'} ) {
	my $f = IO::File->new($pwdfile);
	unless( $f ) {
	    wlog "Failed to open $pwdfile: $!\n";
	    next;
	}
	my @pwd = grep /^$user:/, <$f>;
	close $f;
	next unless @pwd;
	die "Multiple passwords for $user in $pwdfile\n"
	    if @pwd > 1;
	chomp( $pwd = $pwd[0] );
	$pwd =~ s/^.*?://;
	$pwd =~ s/:.*//s;
	die "Account '$user' disabled\n"
	  unless $pwd =~ /^\$1\$/;
	@pwd = split /\$/, $pwd;
	( $seed, $pwd ) = ( $pwd[2], $pwd[3] );
	last;
    }

    my $challenge = md5_base64(localtime().$$.rand(999999));
    print $client "Challenge: $challenge $seed\n";
    my $answerref = $self->headers();

    die "Authentication failed\n"
	unless $pwd and
	    $answerref->{'response'} eq md5_base64( $pwd.$challenge );

    print $client "Ok.\n";

    $self->{'username'} = $user;

    return $pwd;
}


=head1 set_session

Saves current client session.

=cut


sub set_session {
    my ( $self, $secret ) = @_;
    my $client = $self->{'socket'};
    my $ip = $self->{'client_address'};
    my $id = sessionid();

    print $client "SessionID: $id\n\n";

    $secret = substr( md5_base64( $secret.$id ), 0, 8 );
    $secret =~ s/\W+/-/g;
    $id = md5_base64( $secret );

    $self->{'open_sessions'}{"$ip $id"} = 
	[ time(), $self->{'username'}, $secret ];
}


=head1 authenticate

Handles user identification and session retrieval.

=cut


sub authenticate {
    my ( $self, $headers ) = @_;
    my $client = $self->{'socket'};
    my $sid = $headers->{'sessionid'};
    my $ip = $self->{'client_address'};
    my $ses = $self->{'open_sessions'}{"$ip $sid"} || undef;

    if( $ses ) {

	delete $self->{'open_sessions'}{"$ip $sid"};
	
	$self->{'username'} = $ses->[1];

	die "Session expired.\n"
	    if $ses->[0] + $self->{'session_timeout'} < time();

	if( $headers->{'session'} eq 'end' ) {
	    wlog "End session for $self->{'username'}\@$ip\n";
	    print $client "Bye.\n";
	} else {
	    wlog "Continue session for $self->{'username'}\@$ip\n";
	    $self->set_session( $ses->[2] );
	}

    } else {
	
	my $seed = $self->challenge( $headers );
	
	if( $headers->{'session'} eq 'begin' ) {
	    wlog "Begin session for $self->{'username'}\@$ip\n";
	    $self->set_session( $seed ); 
	}
    }
}


### command execution


=head1 manage

Control communication between client and given command.

=cut


sub manage {
    my ( $self, $cmd ) = @_;
    my $client = $self->{'socket'};
    
    my @cmd = split /\s+/, $cmd;
	
    my ( $r, $w, $e );
    my $pid = open3( $w, $r, $e, @cmd );
    $self->{'kill_on_sigpipe'} = $pid; # quit cmd when any pipe closes

    my $err = '';
    my ( $pid2, $pid3 );
    if( $pid2 = fork() ) {
	# fork error reader only if stderr != stdout
        if( $pid3 = fork() ) {
	    # mom: gather errors from command
	    wlog "stderr=$$, stdout=$pid3, stdin=$pid2, cmd=$pid";
	    if( defined($e) ) {
		while( !eof($e) and defined( my $c = getc $e ) ) {
		    $err .= $c;
		    print $client $c;
		    $client->flush();
		}
		wlog "cmd $pid error: $err";
	    }
	} elsif( defined($r) and ( $e != $r ) ) {
	    # pid3: forward from command to client
	    while( !eof($r) and defined( my $c = getc $r ) ) {
		print $client $c;
		$client->flush();
	    }
	    wlog "$$ waiting for $pid" if $self->{'debug'};
	    waitpid $pid, 0;
	    wlog "$$ done." if $self->{'debug'};
	    exit 0;
	}
    } elsif( defined($w) ) {
	# pid2: forward from client to command
	while( <$client> ) {
	    wlog "$$ from client: $_" if $self->{'debug'};
	    print $w $_;
	}
	close $w; # well, how should we tell cmd it's over?
	wlog "$$ done." if $self->{'debug'};
	exit 0;
    }

    # hang around until cmd finished
    wlog "$$ waiting for $pid" if $self->{'debug'};
    while( waitpid($pid,0) != -1 ) {};
    # clean up children.
    wlog "$$ waiting for $pid2" if $self->{'debug'};
    kill 15, $pid2; # ruthlessly kill input from dead listener
    while( waitpid($pid2,0) != -1 ) {};
    wlog "$$ waiting for $pid3" if $self->{'debug'};
    while( waitpid($pid3,0) != -1 ) {}; # nicely wait for output to finish
}


=head1 run_command

Run given command on behalf of current user.

=cut


sub run_command {
    my ( $self, $headers ) = @_;
    my $client = $self->{'socket'} or die "Missing client socket\n";
    my $user = $self->{'username'};
    
    wlog "run $user: $headers->{'command'}\n";
    
    my $cmds = $self->{'users'}{$user};
    die "Bad user\n" unless $cmds;
    
    if( $headers->{'command'} eq 'help' ) {
	print $client "Try these commands for more info:\n";
	foreach( sort keys %$cmds ) {
	    my $n = $_;
	    $n =~ s/\W.*//;
	    print $client "$n\n";
	}
	
    } else {
	
	$ENV{'USER'} = $user;
	
	my $try = $headers->{'command'};
	my $cmd;
	foreach my $re ( keys %$cmds ) {
	    $cmd = $headers->{'command'};
	    if( $cmd =~ /^$re$/ ) {
		my @arg = ( $cmd =~ /^$re$/ );
		$cmd =~ s/^$re$/$cmds->{$re}/;
		$cmd =~ s/\$(\d+)/$arg[$1]/g;
		$cmd =~ s/\$(\{[^\{\}]+\}|\w+)/$ENV{uc($1)}/ge;
		last;
			    }
	    $cmd = undef;
	}

	die "Bad command or illegal parameters. Try 'help'\n" unless $cmd;

	$self->manage( $cmd );
    }
}


=head1 cleanup_sessions

Remove sessions that have timed out.

=cut


sub cleanup_sessions {
    my $self = shift;
    my $list = $self->{'open_sessions'};
    foreach( keys %$list ) {
	delete($list->{$_})
	    if( time() >= $list->{$_}[0] + $self->{'session_timeout'} );
    }
}


### MAIN


=head1 run

Keep listening to and serving connections one by one.

Never returns.

=cut


sub run {
    my $self = shift;

    $self->daemonize() unless $daemon == $self;

    my $reserved = 0;
    while( 1 ) { 

	eval {

	    my $server = 
		IO::Socket::INET->new( 'Proto' => 'tcp',
				       'LocalPort' => $self->{'port'},
				       'Listen' => 3,
				       'Reuse' => 1 );

	    die "Port reserved\n" unless $server;

	    while( my $client = $server->accept() ) { 
		$self->{'socket'} = $client;
		eval {
		    my $ip = $client->peerhost();
		    $ip =~ /^($self->{'addresses'})$/ or 
			die "Bad client address: $ip ($self->{'addresses'})\n";
		    $self->{'client_address'} = $1;

		    $client->autoflush(1);
		    
		    my $headers = $self->headers();
		    
		    $self->authenticate( $headers );
		    
		    $self->run_command( $headers )
			if $headers->{'command'};
		};
		
		if( $@ ) {
		    print $client "Failed: $@\n";
		    wlog "Failed: $@";
		} else {
		    wlog "Success.\n";
		}
		
		print $client "\04";
		$client->flush();
		$client->close();
		$self->{'sigpipe'}=0;
		
	    }
	};
	
	if( $@ =~ /^port reserved/i ) {
	    die $@ if $reserved++ > 4;
	    sleep 1;
	} else {
	    wlog "Connection failed: $@";
	    $reserved = 0;
	}
	
	cleanup_sessions();
    }
}


1;
